# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feature_spec.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'feature_spec.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from databricks.ml_features_common.protos import feature_pb2 as feature__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x66\x65\x61ture_spec.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\x1a\rfeature.proto\"$\n\x14SourceDataColumnInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x9b\x01\n\x11\x46\x65\x61tureColumnInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x14\n\x0c\x66\x65\x61ture_name\x18\x02 \x01(\t\x12\x12\n\nlookup_key\x18\x03 \x03(\t\x12\x13\n\x0boutput_name\x18\x04 \x01(\t\x12\x1c\n\x14timestamp_lookup_key\x18\x05 \x03(\t\x12\x15\n\rdefault_value\x18\x06 \x01(\t\"3\n\x0cInputBinding\x12\x11\n\tparameter\x18\x01 \x01(\t\x12\x10\n\x08\x62ound_to\x18\x02 \x01(\t\"u\n\x12OnDemandColumnInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\x12\x38\n\x0einput_bindings\x18\x02 \x03(\x0b\x32 .featurestorecommon.InputBinding\x12\x13\n\x0boutput_name\x18\x03 \x01(\t\"\xb8\x02\n\nColumnInfo\x12K\n\x17source_data_column_info\x18\x01 \x01(\x0b\x32(.featurestorecommon.SourceDataColumnInfoH\x00\x12\x44\n\x13\x66\x65\x61ture_column_info\x18\x02 \x01(\x0b\x32%.featurestorecommon.FeatureColumnInfoH\x00\x12G\n\x15on_demand_column_info\x18\x03 \x01(\x0b\x32&.featurestorecommon.OnDemandColumnInfoH\x00\x12\x15\n\x07include\x18\n \x01(\x08:\x04true\x12\x11\n\tdata_type\x18\x0b \x01(\t\x12\x1c\n\x14topological_ordering\x18\x0c \x01(\x05\x42\x06\n\x04info\"Q\n\x10\x46\x65\x61tureTableInfo\x12\x12\n\ntable_name\x18\x01 \x01(\t\x12\x10\n\x08table_id\x18\x02 \x01(\t\x12\x17\n\x0flookback_window\x18\x03 \x01(\x01\" \n\x0c\x46unctionInfo\x12\x10\n\x08udf_name\x18\x01 \x01(\t\"\xc5\x02\n\x0b\x46\x65\x61tureSpec\x12\x35\n\rinput_columns\x18\x01 \x03(\x0b\x32\x1e.featurestorecommon.ColumnInfo\x12\x1d\n\x15serialization_version\x18\x02 \x01(\x05\x12\x14\n\x0cworkspace_id\x18\x03 \x01(\x03\x12$\n\x1c\x66\x65\x61ture_store_client_version\x18\x04 \x01(\t\x12:\n\x0cinput_tables\x18\x05 \x03(\x0b\x32$.featurestorecommon.FeatureTableInfo\x12\x39\n\x0finput_functions\x18\x06 \x03(\x0b\x32 .featurestorecommon.FunctionInfo\x12-\n\x08\x66\x65\x61tures\x18\x07 \x03(\x0b\x32\x1b.featurestorecommon.FeatureBu\n.com.databricks.proto.featurestorecommon.jprotoB\x10\x46\x65\x61tureSpecProtoP\x01\xa0\x01\x01\xe2?+\n\'com.databricks.proto.featurestorecommon\x10\x01')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_spec_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.com.databricks.proto.featurestorecommon.jprotoB\020FeatureSpecProtoP\001\240\001\001\342?+\n\'com.databricks.proto.featurestorecommon\020\001'
  _globals['_SOURCEDATACOLUMNINFO']._serialized_start=80
  _globals['_SOURCEDATACOLUMNINFO']._serialized_end=116
  _globals['_FEATURECOLUMNINFO']._serialized_start=119
  _globals['_FEATURECOLUMNINFO']._serialized_end=274
  _globals['_INPUTBINDING']._serialized_start=276
  _globals['_INPUTBINDING']._serialized_end=327
  _globals['_ONDEMANDCOLUMNINFO']._serialized_start=329
  _globals['_ONDEMANDCOLUMNINFO']._serialized_end=446
  _globals['_COLUMNINFO']._serialized_start=449
  _globals['_COLUMNINFO']._serialized_end=761
  _globals['_FEATURETABLEINFO']._serialized_start=763
  _globals['_FEATURETABLEINFO']._serialized_end=844
  _globals['_FUNCTIONINFO']._serialized_start=846
  _globals['_FUNCTIONINFO']._serialized_end=878
  _globals['_FEATURESPEC']._serialized_start=881
  _globals['_FEATURESPEC']._serialized_end=1206
# @@protoc_insertion_point(module_scope)
