# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: feature_store_serving.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'feature_store_serving.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mlflow.protos.scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x65\x61ture_store_serving.proto\x12\x12\x66\x65\x61turestorecommon\x1a\x15scalapb/scalapb.proto\"\x9b\x04\n\x15OnlineStoreForServing\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x1a\n\x12read_secret_prefix\x18\x03 \x01(\t\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x04 \x01(\x03\x12\x33\n\nmysql_conf\x18\x05 \x01(\x0b\x32\x1d.featurestorecommon.MySqlConfH\x00\x12<\n\x0fsql_server_conf\x18\x06 \x01(\x0b\x32!.featurestorecommon.SqlServerConfH\x00\x12\x39\n\rdynamodb_conf\x18\x07 \x01(\x0b\x32 .featurestorecommon.DynamoDbConfH\x00\x12\x39\n\rcosmosdb_conf\x18\x08 \x01(\x0b\x32 .featurestorecommon.CosmosDbConfH\x00\x12=\n\x0f\x62rickstore_conf\x18\t \x01(\x0b\x32\".featurestorecommon.BrickstoreConfH\x00\x12\x31\n\nquery_mode\x18\x15 \x01(\x0e\x32\x1d.featurestorecommon.QueryModeB\x0f\n\rextra_configs\"\'\n\tMySqlConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"+\n\rSqlServerConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"\x1e\n\x0c\x44ynamoDbConf\x12\x0e\n\x06region\x18\x01 \x01(\t\"#\n\x0c\x43osmosDbConf\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\"^\n\x0e\x42rickstoreConf\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x19\n\x11table_serving_url\x18\x03 \x01(\t\x12\x15\n\rinstance_name\x18\x04 \x01(\t\"j\n\x0e\x46\x65\x61tureDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x19\n\x11\x64\x61ta_type_details\x18\x03 \x01(\t\"R\n\x11PrimaryKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"T\n\x13TimestampKeyDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\tdata_type\x18\x02 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xe2\x02\n\x12OnlineFeatureTable\x12\x1a\n\x12\x66\x65\x61ture_table_name\x18\x01 \x01(\t\x12!\n\x19online_feature_table_name\x18\x02 \x01(\t\x12?\n\x0conline_store\x18\x03 \x01(\x0b\x32).featurestorecommon.OnlineStoreForServing\x12;\n\x0cprimary_keys\x18\x04 \x03(\x0b\x32%.featurestorecommon.PrimaryKeyDetails\x12\x18\n\x10\x66\x65\x61ture_table_id\x18\x05 \x01(\t\x12\x34\n\x08\x66\x65\x61tures\x18\x06 \x03(\x0b\x32\".featurestorecommon.FeatureDetails\x12?\n\x0etimestamp_keys\x18\x07 \x03(\x0b\x32\'.featurestorecommon.TimestampKeyDetails\"\xa4\x01\n\x17\x46\x65\x61tureTablesForServing\x12=\n\ronline_tables\x18\x01 \x03(\x0b\x32&.featurestorecommon.OnlineFeatureTable\x12J\n\x17\x66\x65\x61ture_definition_type\x18\x02 \x01(\x0e\x32).featurestorecommon.FeatureDefinitionType\"+\n\rMySqlMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"/\n\x11SqlServerMetadata\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"B\n\x10\x44ynamoDbMetadata\x12\x0e\n\x06region\x18\x01 \x01(\t\x12\x11\n\ttable_arn\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"K\n\x10\x43osmosDbMetadata\x12\x13\n\x0b\x61\x63\x63ount_uri\x18\x01 \x01(\t\x12\x15\n\rcontainer_uri\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\"\xa9\x04\n\x0bOnlineStore\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12;\n\x0emysql_metadata\x18\n \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\x0b \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0c \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\r \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"\xdf\x04\n\x13OnlineStoreDetailed\x12(\n\x05\x63loud\x18\x01 \x01(\x0e\x32\x19.featurestorecommon.Cloud\x12\x31\n\nstore_type\x18\x02 \x01(\x0e\x32\x1d.featurestorecommon.StoreType\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x04 \x01(\x03\x12\x1e\n\x16last_updated_timestamp\x18\x05 \x01(\x03\x12\x12\n\ncreator_id\x18\x06 \x01(\t\x12\x10\n\x04host\x18\x07 \x01(\tB\x02\x18\x01\x12\x10\n\x04port\x18\x08 \x01(\x05\x42\x02\x18\x01\x12\x1b\n\x13last_update_user_id\x18\t \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\n \x03(\t\x12\x1a\n\x12read_secret_prefix\x18\x0b \x01(\t\x12;\n\x0emysql_metadata\x18\x0c \x01(\x0b\x32!.featurestorecommon.MySqlMetadataH\x00\x12\x44\n\x13sql_server_metadata\x18\r \x01(\x0b\x32%.featurestorecommon.SqlServerMetadataH\x00\x12\x41\n\x11\x64ynamodb_metadata\x18\x0e \x01(\x0b\x32$.featurestorecommon.DynamoDbMetadataH\x00\x12\x41\n\x11\x63osmosdb_metadata\x18\x0f \x01(\x0b\x32$.featurestorecommon.CosmosDbMetadataH\x00\x42\x15\n\x13\x61\x64\x64itional_metadata\"\x83\x01\n\x1c\x46\x65\x61tureFunctionParameterInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\ttype_text\x18\x02 \x01(\t\x12\x11\n\ttype_json\x18\x03 \x01(\t\x12/\n\ttype_name\x18\x04 \x01(\x0e\x32\x1c.featurestorecommon.DataType\"\xdb\x01\n\x19\x46\x65\x61tureFunctionForServing\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\x46\n\x0cinput_params\x18\x02 \x03(\x0b\x32\x30.featurestorecommon.FeatureFunctionParameterInfo\x12/\n\tdata_type\x18\x03 \x01(\x0e\x32\x1c.featurestorecommon.DataType\x12\x16\n\x0e\x66ull_data_type\x18\x04 \x01(\t\x12\x1a\n\x12routine_definition\x18\x05 \x01(\t\"c\n\x1a\x46\x65\x61tureFunctionsForServing\x12\x45\n\x0e\x66\x66_for_serving\x18\x01 \x03(\x0b\x32-.featurestorecommon.FeatureFunctionForServing*$\n\x05\x43loud\x12\x07\n\x03\x41WS\x10\x01\x12\t\n\x05\x41ZURE\x10\x02\x12\x07\n\x03GCP\x10\x03*\x81\x01\n\tStoreType\x12\x10\n\x0c\x41URORA_MYSQL\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\t\n\x05MYSQL\x10\x03\x12\x0c\n\x08\x44YNAMODB\x10\x04\x12\x0c\n\x08\x43OSMOSDB\x10\x05\x12\x0e\n\nBRICKSTORE\x10\x06\x12\x1b\n\x17\x44\x41TABRICKS_ONLINE_STORE\x10\x07*\xae\x01\n\x08\x44\x61taType\x12\x0b\n\x07INTEGER\x10\x01\x12\t\n\x05\x46LOAT\x10\x02\x12\x0b\n\x07\x42OOLEAN\x10\x03\x12\n\n\x06STRING\x10\x04\x12\n\n\x06\x44OUBLE\x10\x05\x12\x08\n\x04LONG\x10\x06\x12\r\n\tTIMESTAMP\x10\x07\x12\x08\n\x04\x44\x41TE\x10\x08\x12\t\n\x05SHORT\x10\t\x12\t\n\x05\x41RRAY\x10\n\x12\x07\n\x03MAP\x10\x0b\x12\n\n\x06\x42INARY\x10\x0c\x12\x0b\n\x07\x44\x45\x43IMAL\x10\r\x12\n\n\x06STRUCT\x10\x0e*4\n\tQueryMode\x12\x16\n\x12PRIMARY_KEY_LOOKUP\x10\x00\x12\x0f\n\x0bRANGE_QUERY\x10\x01*m\n\x15\x46\x65\x61tureDefinitionType\x12\'\n#FEATURE_DEFINITION_TYPE_UNSPECIFIED\x10\x00\x12\x11\n\rFEATURE_TABLE\x10\x01\x12\x18\n\x14\x44\x45\x43LARATIVE_FEATURES\x10\x02\x42}\n.com.databricks.proto.featurestorecommon.jprotoB\x18\x46\x65\x61tureStoreServingProtoP\x01\xa0\x01\x01\xe2?+\n\'com.databricks.proto.featurestorecommon\x10\x01')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_store_serving_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.com.databricks.proto.featurestorecommon.jprotoB\030FeatureStoreServingProtoP\001\240\001\001\342?+\n\'com.databricks.proto.featurestorecommon\020\001'
  _globals['_ONLINESTORE'].fields_by_name['host']._loaded_options = None
  _globals['_ONLINESTORE'].fields_by_name['host']._serialized_options = b'\030\001'
  _globals['_ONLINESTORE'].fields_by_name['port']._loaded_options = None
  _globals['_ONLINESTORE'].fields_by_name['port']._serialized_options = b'\030\001'
  _globals['_ONLINESTOREDETAILED'].fields_by_name['host']._loaded_options = None
  _globals['_ONLINESTOREDETAILED'].fields_by_name['host']._serialized_options = b'\030\001'
  _globals['_ONLINESTOREDETAILED'].fields_by_name['port']._loaded_options = None
  _globals['_ONLINESTOREDETAILED'].fields_by_name['port']._serialized_options = b'\030\001'
  _globals['_CLOUD']._serialized_start=3531
  _globals['_CLOUD']._serialized_end=3567
  _globals['_STORETYPE']._serialized_start=3570
  _globals['_STORETYPE']._serialized_end=3699
  _globals['_DATATYPE']._serialized_start=3702
  _globals['_DATATYPE']._serialized_end=3876
  _globals['_QUERYMODE']._serialized_start=3878
  _globals['_QUERYMODE']._serialized_end=3930
  _globals['_FEATUREDEFINITIONTYPE']._serialized_start=3932
  _globals['_FEATUREDEFINITIONTYPE']._serialized_end=4041
  _globals['_ONLINESTOREFORSERVING']._serialized_start=75
  _globals['_ONLINESTOREFORSERVING']._serialized_end=614
  _globals['_MYSQLCONF']._serialized_start=616
  _globals['_MYSQLCONF']._serialized_end=655
  _globals['_SQLSERVERCONF']._serialized_start=657
  _globals['_SQLSERVERCONF']._serialized_end=700
  _globals['_DYNAMODBCONF']._serialized_start=702
  _globals['_DYNAMODBCONF']._serialized_end=732
  _globals['_COSMOSDBCONF']._serialized_start=734
  _globals['_COSMOSDBCONF']._serialized_end=769
  _globals['_BRICKSTORECONF']._serialized_start=771
  _globals['_BRICKSTORECONF']._serialized_end=865
  _globals['_FEATUREDETAILS']._serialized_start=867
  _globals['_FEATUREDETAILS']._serialized_end=973
  _globals['_PRIMARYKEYDETAILS']._serialized_start=975
  _globals['_PRIMARYKEYDETAILS']._serialized_end=1057
  _globals['_TIMESTAMPKEYDETAILS']._serialized_start=1059
  _globals['_TIMESTAMPKEYDETAILS']._serialized_end=1143
  _globals['_ONLINEFEATURETABLE']._serialized_start=1146
  _globals['_ONLINEFEATURETABLE']._serialized_end=1500
  _globals['_FEATURETABLESFORSERVING']._serialized_start=1503
  _globals['_FEATURETABLESFORSERVING']._serialized_end=1667
  _globals['_MYSQLMETADATA']._serialized_start=1669
  _globals['_MYSQLMETADATA']._serialized_end=1712
  _globals['_SQLSERVERMETADATA']._serialized_start=1714
  _globals['_SQLSERVERMETADATA']._serialized_end=1761
  _globals['_DYNAMODBMETADATA']._serialized_start=1763
  _globals['_DYNAMODBMETADATA']._serialized_end=1829
  _globals['_COSMOSDBMETADATA']._serialized_start=1831
  _globals['_COSMOSDBMETADATA']._serialized_end=1906
  _globals['_ONLINESTORE']._serialized_start=1909
  _globals['_ONLINESTORE']._serialized_end=2462
  _globals['_ONLINESTOREDETAILED']._serialized_start=2465
  _globals['_ONLINESTOREDETAILED']._serialized_end=3072
  _globals['_FEATUREFUNCTIONPARAMETERINFO']._serialized_start=3075
  _globals['_FEATUREFUNCTIONPARAMETERINFO']._serialized_end=3206
  _globals['_FEATUREFUNCTIONFORSERVING']._serialized_start=3209
  _globals['_FEATUREFUNCTIONFORSERVING']._serialized_end=3428
  _globals['_FEATUREFUNCTIONSFORSERVING']._serialized_start=3430
  _globals['_FEATUREFUNCTIONSFORSERVING']._serialized_end=3529
# @@protoc_insertion_point(module_scope)
